import os
import numpy as np
from PIL import Image
import random
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt

class CASIA(object):
	"""
		Reads the CASIA dataset (uses train, test and validation set)
	"""
	
	def __init__(self, data_dir, image_size=(250, 250), training_data_fraction=0.7, validation_data_fraction=0.15):
		"""
			__init__(CASIA, str, image_size, float) -> None
		"""
		self.training_faces = []
		self.test_faces = []
		self.validation_faces = []
		self.num_training_faces = 0
		self.num_test_faces = 0
		self.num_validation_faces = 0
		self.data_dir = data_dir
		self.image_size = image_size
		
		# Read the labels of faces to be used for test and training
		self.read_faces(training_data_fraction, validation_data_fraction)
	
	
	def read_faces(self, training_data_fraction, validation_data_fraction):
		"""
			read_faces(CASIA, float, float) -> None
			Reads the faces and splits the data into test, validation and training set
		"""
		# Read the label for all faces
		faces = []
		for _, faces, _ in os.walk(self.data_dir):
			break
		
		# Calculate the number of faces
		num_faces = len(faces)
		self.num_training_faces = int(num_faces * training_data_fraction)
		self.num_validation_faces = int(num_faces * validation_data_fraction)
		self.num_test_faces = num_faces - (self.num_training_faces + self.num_validation_faces)
		
		# Divide the data into test and train set
		random.shuffle(faces)
		self.training_faces = faces[:self.num_training_faces]
		self.validation_faces = faces[self.num_training_faces: self.num_training_faces+self.num_validation_faces]
		self.test_faces = faces[self.num_training_faces+self.num_validation_faces:]
		
		print self.training_faces[1:10]
		print self.test_faces[1:10]
		print self.validation_faces[1:10]
		
		# Read the training faces
		j = 0
		for face in self.training_faces:
			# Read all the filenames
			filenames = []
			for _, _, filenames in os.walk(os.path.join(self.data_dir, face)):
				break
			
			# Read all faces
			cur_faces = np.zeros((len(filenames), self.image_size[0], self.image_size[1], 3))
			k = 0			
			n = len(filenames)
			print "Reading Face:", j+1, "of", self.num_training_faces, "Progress: ", \
										"{:.3f}".format((float(k)/n) * 100), "%",
			for filename in filenames:
				cur_faces[k, :, :, :] = self.load_image(os.path.join(self.data_dir, face, filename))
				k += 1
				print "\rReading Face:", j+1, "of", self.num_training_faces, "Progress: ", \
											"{:.3f}".format((float(k)/n) * 100), "%",
			
			print ''
			# Save the faces
			self.training_faces[j] = cur_faces
			j += 1


		# Read the validation faces
		j = 0
		for face in self.validation_faces:
			# Read all the filenames
			filenames = []
			for _, _, filenames in os.walk(os.path.join(self.data_dir, face)):
				break
			
			# Read all faces
			cur_faces = np.zeros((len(filenames), self.image_size[0], self.image_size[1], 3))
			k = 0
			n = len(filenames)
			print "Reading Face:", j+1, "of", self.num_validation_faces, "Progress: ", \
										"{:.3f}".format((float(k)/n) * 100), "%",
			for filename in filenames:
				cur_faces[k, :, :, :] = self.load_image(os.path.join(self.data_dir, face, filename))
				k += 1
				print "\rReading Face:", j+1, "of", self.num_validation_faces, "Progress: ", \
											"{:.3f}".format((float(k)/n) * 100), "%",
			
			print ''
			# Save the faces
			self.validation_faces[j] = cur_faces
			j += 1


		# Read the test faces
		j = 0
		for face in self.test_faces:
			# Read all the filenames
			filenames = []
			for _, _, filenames in os.walk(os.path.join(self.data_dir, face)):
				break
			
			# Read all faces
			cur_faces = np.zeros((len(filenames), self.image_size[0], self.image_size[1], 3))
			k = 0
			n = len(filenames)
			print "Reading Face:", j+1, "of", self.num_test_faces, "Progress: ", \
										"{:.3f}".format((float(k)/n) * 100), "%",
			for filename in filenames:
				cur_faces[k, :, :, :] = self.load_image(os.path.join(self.data_dir, face, filename))
				k += 1
				print "\rReading Face:", j+1, "of", self.num_test_faces, "Progress: ", \
											"{:.3f}".format((float(k)/n) * 100), "%",
			
			print ''
			# Save the faces
			self.test_faces[j] = cur_faces
			j += 1
		
	
	
	
	def load_image(self, filename):
		"""
			load_image(CASIA, str) -> ndarray
			Reads the image from the specified file and returns an image of size = self.image_size
		"""
		# Read the image
		im = Image.open(filename)
		im.load()
				
		# Resize the image
		im = im.resize(self.image_size, Image.LANCZOS)
		
		# Make the image a 3 channel image
		im = np.asarray(im)
		shape = np.shape(im)
		if len(shape) == 2:	# Only 1 channel
			im = np.stack([im, im, im], axis=2)	# Convert to a 3 channel image
		elif shape[2] == 4:	# 4 Channels
			im = im[:, :, 1:4]	# Keep the RGB channels
		
		# Bring the image in range [0, 1]
		im = np.ndarray.astype(im, np.float32) / 255
		
		return im
		
	
	
	def get_data(self, batch_size, is_validation=False):
		"""
			get_data(CASIA, int, bool) -> (images1, images2, is_similar)
			Returns a batch to the client
		"""
		# Initialize the data
		images1 = np.zeros((batch_size, self.image_size[0], self.image_size[1], 3))
		images2 = np.zeros((batch_size, self.image_size[0], self.image_size[1], 3))
		is_similar = np.zeros((batch_size, 1))
		
		# Read the batch
		similar = True		
		for i in range(batch_size):
			im1 = None
			im2 = None
			is_sim = 0.0
			if similar:
				im1, im2 = self.read_similar(is_validation)
				is_sim = 1.0
				similar = False
			else:
				im1, im2 = self.read_different(is_validation)
				similar = True
			
			images1[i, :, :, :] = im1
			images2[i, :, :, :] = im2
			is_similar[i] = is_sim
		
		return (images1, images2, is_similar)
		
		
	
	def read_similar(self, is_validation):
		"""
			read_similar(CASIA, bool) -> (image1, image2)
			Reads two images of a randomly chosen face
		"""
		if is_validation:
			face_id = np.random.randint(0, self.num_validation_faces)
			num_faces = np.shape(self.validation_faces[face_id])[0]
			im_ids = np.random.randint(0, num_faces, size=(2))
			while im_ids[0] == im_ids[1]:
				im_ids = np.random.randint(0, num_faces, size=(2))
			
			im1 = self.validation_faces[face_id][im_ids[0], :, :, :]
			im2 = self.validation_faces[face_id][im_ids[1], :, :, :]
			return (im1, im2)
			
		else:
			face_id = np.random.randint(0, self.num_training_faces)
			num_faces = np.shape(self.training_faces[face_id])[0]
			im_ids = np.random.randint(0, num_faces, size=(2))
			while im_ids[0] == im_ids[1]:
				im_ids = np.random.randint(0, num_faces, size=(2))
			
			im1 = self.training_faces[face_id][im_ids[0], :, :, :]
			im2 = self.training_faces[face_id][im_ids[1], :, :, :]
			return (im1, im2)
	
	
		
	def read_different(self, is_validation):
		"""
			read_similar(CASIA, bool) -> (image1, image2)
			Reads two images from a randomly chosen different faces
		"""
		if is_validation:
			face_ids = np.random.randint(0, self.num_validation_faces, size=(2))
			while face_ids[0] == face_ids[1]:
				face_ids = np.random.randint(0, self.num_validation_faces, size=(2))
			
			num_faces1 = np.shape(self.validation_faces[face_ids[0]])[0]
			num_faces2 = np.shape(self.validation_faces[face_ids[1]])[0]
			
			im_id1 = np.random.randint(0, num_faces1)
			im_id2 = np.random.randint(0, num_faces2)
			
			im1 = self.validation_faces[face_ids[0]][im_id1, :, :, :]
			im2 = self.validation_faces[face_ids[1]][im_id2, :, :, :]
			return (im1, im2)
			
		else:
			face_ids = np.random.randint(0, self.num_training_faces, size=(2))
			while face_ids[0] == face_ids[1]:
				face_ids = np.random.randint(0, self.num_training_faces, size=(2))
			
			num_faces1 = np.shape(self.training_faces[face_ids[0]])[0]
			num_faces2 = np.shape(self.training_faces[face_ids[1]])[0]
			
			im_id1 = np.random.randint(0, num_faces1)
			im_id2 = np.random.randint(0, num_faces2)
			
			im1 = self.training_faces[face_ids[0]][im_id1, :, :, :]
			im2 = self.training_faces[face_ids[1]][im_id2, :, :, :]
			return (im1, im2)
		
		
	
	def get_data_test(self, batch_size):
		"""
			get_data_test(CASIA, int) -> (images1, images2, is_similar)
			Returns a batch to the client from test set
		"""
		# Initialize the data
		images1 = np.zeros((batch_size, self.image_size[0], self.image_size[1], 3))
		images2 = np.zeros((batch_size, self.image_size[0], self.image_size[1], 3))
		is_similar = np.zeros((batch_size, 1))
		
		# Read the batch
		similar = True		
		for i in range(batch_size):
			im1 = None
			im2 = None
			is_sim = 0.0
			if similar:
				im1, im2 = self.read_similar_test()
				is_sim = 1.0
				similar = False
			else:
				im1, im2 = self.read_different_test()
				similar = True
			
			images1[i, :, :, :] = im1
			images2[i, :, :, :] = im2
			is_similar[i] = is_sim
		
		return (images1, images2, is_similar)
		
	
	
	def read_similar_test(self):
		"""
			read_similar_test(CASIA) -> (image1, image2)
			Reads two images of a randomly chosen face from test set
		"""
		face_id = np.random.randint(0, self.num_test_faces)
		num_faces = np.shape(self.test_faces[face_id])[0]
		im_ids = np.random.randint(0, num_faces, size=(2))
		while im_ids[0] == im_ids[1]:
			im_ids = np.random.randint(0, num_faces, size=(2))
		
		im1 = self.test_faces[face_id][im_ids[0], :, :, :]
		im2 = self.test_faces[face_id][im_ids[1], :, :, :]
		return (im1, im2)
	
	
		
	def read_different_test(self):
		"""
			read_different_test(CASIA) -> (image1, image2)
			Reads two images from randomly chosen different faces from test set
		"""
		face_ids = np.random.randint(0, self.num_test_faces, size=(2))
		while face_ids[0] == face_ids[1]:
			face_ids = np.random.randint(0, self.num_test_faces, size=(2))
		
		num_faces1 = np.shape(self.test_faces[face_ids[0]])[0]
		num_faces2 = np.shape(self.test_faces[face_ids[1]])[0]
		
		im_id1 = np.random.randint(0, num_faces1)
		im_id2 = np.random.randint(0, num_faces2)
		
		im1 = self.test_faces[face_ids[0]][im_id1, :, :, :]
		im2 = self.test_faces[face_ids[1]][im_id2, :, :, :]
		return (im1, im2)
	
			
		
	def show_image(self, image, image2=None, is_similar=0):
		if image2 == None:
			plt.imshow(image)
			plt.show()
		else:
			plt.imshow(np.concatenate([image, image2], axis=1))
			print is_similar
			plt.show()
		
		
if __name__ == '__main__':
	casia = CASIA('../Datasets/CASIA/CASIA-WebFace', (32, 32))
	images1, images2, is_similar = casia.get_data(128)
	images1, images2, is_similar = casia.get_data(128, is_test=True)
