"""
    File implementing a wideresnet based Siamese neural network for verification
    task on CASIA 
"""



import tensorflow as tf
import numpy as np
from read_casia_in_mem_v2 import *
from keras import backend as K
from keras.layers import Dense, Activation, Convolution2D, BatchNormalization, Dropout, Merge
from keras.regularizers import l2
import math
import random

# Useful variables
image_size = 64
batch_size = 128
learning_rate = 1e-4
num_batches = 100000
wrn_k = 2
wrn_n = 4
reg_param = 0.00001
prev_best_val_cost = float('inf')
prev_val_cost = float('inf')
inc_cost_count = 0
same_cost_count = 0
inc_cost_threshold = 5
same_cost_threshold = 5
eps = 1e-3

random.seed(100)

def residual_block(x, first=False, increase_dim=False, nb_filters=16):
	if increase_dim:
		first_stride = (2, 2)
	else:
		first_stride = (1, 1)
	
	if first:
		bn_pre_relu = x
	else:
		bn_pre_conv = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(x)
		bn_pre_relu = Activation('relu')(bn_pre_conv)
	
	
	# First layer
	l = Convolution2D(nb_filter=nb_filters, nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
									subsample=first_stride, W_regularizer=l2(reg_param))(bn_pre_relu)
	l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
	l = Activation('relu')(l)
	
	# Dropout
	l = Dropout(0.3)(l)
	
	# Second layer
	conv2 = Convolution2D(nb_filter=nb_filters, nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
									subsample=(1, 1), W_regularizer=l2(reg_param))(l)
	
	
	if increase_dim:
		proj = Convolution2D(nb_filter=nb_filters, nb_row=1, nb_col=1, init='he_normal', border_mode='same', \
									subsample=(1, 1), W_regularizer=l2(reg_param), bias=False)(conv2)
		l = conv2 + proj
	elif first:
		proj = Convolution2D(nb_filter=nb_filters, nb_row=1, nb_col=1, init='he_normal', border_mode='same', \
									subsample=(1, 1), W_regularizer=l2(reg_param), bias=False)(conv2)
		l = conv2 + proj
	else:
		l = conv2 + x
	
	return l



inputs1 = tf.placeholder(dtype=tf.float32, shape=[None, image_size, image_size, 3])
inputs2 = tf.placeholder(dtype=tf.float32, shape=[None, image_size, image_size, 3])
labels = tf.placeholder(dtype=tf.float32, shape=[None, 1])

################################################ Run WideResNet ##################################################
# Merge inputs1 and inputs2
inputs = tf.concat(0, [inputs1, inputs2])

n_filters = [16, 16*wrn_k, 32*wrn_k]

# First layer
l = Convolution2D(nb_filter=n_filters[0], nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
									subsample=(1, 1), W_regularizer=l2(reg_param))(inputs)
l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
l = Activation('relu')(l)

# Second layer, a stack of residual blocks
l = residual_block(l, first=True, nb_filters=n_filters[1])
for _ in range(1, wrn_n):
	l = residual_block(l, nb_filters=n_filters[1])

# Third layer, a stack of residual blocks
l = residual_block(l, increase_dim=True, nb_filters=n_filters[2])
for _ in range(1, wrn_n+2):
	l = residual_block(l, nb_filters=n_filters[2])

l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
final_layer_inputs = Activation('relu')(l)

#######NEW#########
final_layer_inputs = tf.contrib.layers.flatten(final_layer_inputs)
final_layer_inputs = Dense(400, init='he_normal', activation='relu', W_regularizer=l2(reg_param))(final_layer_inputs)
final_layer_inputs = Dense(100, init='he_normal', activation='relu', W_regularizer=l2(reg_param))(final_layer_inputs)
final_layer_inputs1 = final_layer_inputs[:batch_size, :]
final_layer_inputs2 = final_layer_inputs[batch_size:, :]
#######NEW#########

# Divide inputs to final layer
#final_layer_inputs1 = tf.contrib.layers.flatten(final_layer_inputs[:batch_size, :, :, :])
#final_layer_inputs2 = tf.contrib.layers.flatten(final_layer_inputs[batch_size:, :, :, :])


########################################## Run the Comparator #####################################################
output = tf.abs(final_layer_inputs1 - final_layer_inputs2)
output = Dense(1, init='he_normal', activation='sigmoid', W_regularizer=l2(reg_param))(output)

network_cost = -tf.reduce_mean(labels*tf.log(output) + (1.0-labels)*tf.log(1.0-output))
#"""

train_step = tf.train.AdamOptimizer(learning_rate).minimize(network_cost)
accuracy = tf.reduce_mean(tf.cast(tf.equal(tf.round(output), labels), tf.float32))



def train(images1, images2, is_similar):
	"""
		Train the network for verification task
	"""
	_, batch_cost = sess.run((train_step, network_cost), feed_dict={inputs1: images1, inputs2: images2, labels: is_similar, K.learning_phase(): 1})
	return batch_cost
	
	
def calculate_accuracy(images1, images2, is_similar, is_learning=0):
	"""
		Calculate accuracy for verification task
	"""
	#print sess.run(output, feed_dict={inputs1: images1, inputs2: images2, labels: is_similar})
	return sess.run(accuracy, feed_dict={inputs1: images1, inputs2: images2, labels: is_similar, K.learning_phase(): is_learning})
	
	

sess = tf.InteractiveSession()
saver = tf.train.Saver()
if os.path.isfile('./SiameseModel/model.meta'):
	saver.restore(sess, './SiameseModel/model')
else:
	sess.run(tf.global_variables_initializer())

K.set_session(sess)

# Useful controls
display_step = 50
save_step = 1000



casia = CASIA('../Datasets/CASIA/CASIA-WebFace', (image_size, image_size))


#"""
for i in range(num_batches):
	images1, images2, is_similar = casia.get_data(batch_size)
	
	# Train on the batch
	train_cost = train(images1, images2, is_similar)	
	#temp = sess.run(output, feed_dict={inputs1: images1, inputs2: images2, labels: is_similar, K.learning_phase(): 0})
	#print np.concatenate([temp, is_similar, np.round(temp)], axis=1)
		
	if i % display_step == 0:		
		# Calculate accuracy on Validation Images
		images1, images2, is_similar = casia.get_data(batch_size, is_validation=True)
		val_accuracy = calculate_accuracy(images1, images2, is_similar)
				
		# Fetch some training images and calculate accuracy
		images1, images2, is_similar = casia.get_data(batch_size)
		train_accuracy = calculate_accuracy(images1, images2, is_similar)	
		
		print "Batch:", (i+1), "Validation Accuracy:", val_accuracy, "Train Cost:", train_cost, "Train Accuracy:", train_accuracy
	
	if i % save_step == 0:
		# Approximate validation cost over 
		val_cost = 0.0
		val_accuracy = 0.0
		for j in range(100):	
			images1, images2, is_similar = casia.get_data(batch_size, is_validation=True)
			v_cost, v_acc = sess.run((network_cost, accuracy), feed_dict={inputs1: images1, inputs2:images2, labels: is_similar, K.learning_phase(): 0})
			val_cost += v_cost
			val_accuracy += v_acc
		val_cost /= 100
		val_accuracy /= 100
		print 'Validation Set - Cost:', val_cost, 'Accuracy:', val_accuracy
		
		if val_cost < prev_best_val_cost:
			prev_best_val_cost = val_cost
			saver.save(sess, './SiameseModel_Best/model')	
		
		if val_cost - prev_val_cost >= eps:
			inc_cost_count += 1
			same_cost_count = 0
		elif val_cost - prev_val_cost <= -eps:
			inc_cost_count = 0
			same_cost_count = 0
		else:
			inc_cost_count = 0
			same_cost_count += 1		
		
		prev_val_cost = val_cost
		
		saver.save(sess, './SiameseModel/model')
		
		if inc_cost_count >= inc_cost_threshold:
			print 'Ending Training: Validation Cost Increasing since', inc_cost_threshold*1000, 'iterations'
			break
		if same_cost_count >= same_cost_threshold:			
			print 'Ending Training: Validation Cost Same since', same_cost_threshold*1000, 'iterations'
			break
		

#"""

# Calculate Test Accuracy
"""
num_trials = 1000
test_acc = 0.0
for i in range(num_trials):
	images1, images2, is_similar = casia.get_data_test(batch_size)
	test_accuracy = calculate_accuracy(images1, images2, is_similar)
	test_acc += test_accuracy
	print "Batch:", (i+1), "Test Accuracy:", test_accuracy

print "Accuracy:", (test_acc/num_trials)	
"""
