"""
    File implementing wideresnet based convarc for the verification task on 
    CASIA. Includes code for training, testing and showing attention patches.
"""



import tensorflow as tf
import numpy as np
from read_casia_in_mem_v2 import *
from keras import backend as K
from keras.layers import Dense, Activation, Convolution2D, BatchNormalization, Dropout, Merge
from keras.regularizers import l2
import math
import random
import matplotlib.patches as patches
from PIL import Image

# Useful variables
image_size = 64
attention_window_size = 4
num_glimpses = 8
batch_size = 128
learning_rate = 1e-4
hidden_layer_size = 256
num_batches = 100000
forget_bias = 0.2
wrn_k = 2
wrn_n = 4
reg_param = 0.0
prev_best_val_cost = float('inf')
prev_val_cost = float('inf')
inc_cost_count = 0
same_cost_count = 0
inc_cost_threshold = 3
same_cost_threshold = 5
eps = 1e-3
idx_to_show=73

random.seed(100)


def get_glimpse_batch(images, h, W_g, idx):
	"""
		Get glimpse from the batch based on the hidden state
	"""
	omega = tf.nn.tanh(tf.matmul(W_g, h))
	x_hat = omega[0, :]
	y_hat = omega[1, :]
	delta_hat = omega[2, :]
	image_size = 32		# Change this to 16 if using image_size = 32 globally
	num_input_channels = 32*wrn_k
	
	x = tf.reshape((image_size - 1) * (x_hat + 1) / 2, (-1, 1))
	y = tf.reshape((image_size - 1) * (y_hat + 1) / 2, (-1, 1))
	delta = tf.reshape((image_size / float(attention_window_size)) * (1 - tf.abs(delta_hat)), (-1, 1))
	gamma = tf.reshape(tf.exp(1 - 2*tf.abs(delta_hat)), (-1, 1))
	
	patch = [y[idx, 0], x[idx, 0], delta[idx, 0]]

	grid_i = tf.reshape(tf.cast(tf.range(attention_window_size), tf.float32), [1, -1])
	ux = x + (grid_i - attention_window_size/2.0 - 0.5) * delta
	uy = y + (grid_i - attention_window_size/2.0 - 0.5) * delta

	a = tf.reshape(tf.cast(tf.range(image_size), tf.float32), [1, 1, -1])
	b = tf.reshape(tf.cast(tf.range(image_size), tf.float32), [1, 1, -1])
	ux = tf.reshape(ux, [-1, attention_window_size, 1])
	uy = tf.reshape(uy, [-1, attention_window_size, 1])
	pi = tf.constant(math.pi, dtype=tf.float32)

	gamma = tf.reshape(gamma, (-1, 1, 1))
	Fx = 1.0 / (tf.mul(pi, gamma) * (1.0 + tf.square((a - ux) / gamma)))
	Fy = 1.0 / (tf.mul(pi, gamma) * (1.0 + tf.square((b - uy) / gamma)))

	Fx = Fx / tf.reduce_sum(Fx, 2, keep_dims=True)
	Fy = Fy / tf.reduce_sum(Fy, 2, keep_dims=True)
	
	Fx = tf.stack([Fx]*num_input_channels, axis=1)
	Fy = tf.stack([Fy]*num_input_channels, axis=1)
	
	return (tf.batch_matmul(Fy, tf.batch_matmul(tf.transpose(images, perm=[0, 3, 1, 2]), \
											tf.transpose(Fx, perm=[0, 1, 3, 2]))), patch)
											
											
											
def residual_block(x, first=False, increase_dim=False, nb_filters=16):
	if increase_dim:
		first_stride = (2, 2)
	else:
		first_stride = (1, 1)
	
	if first:
		bn_pre_relu = x
	else:
		bn_pre_conv = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(x)
		bn_pre_relu = Activation('relu')(bn_pre_conv)
	
	
	# First layer
	l = Convolution2D(nb_filter=nb_filters, nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
									subsample=first_stride, W_regularizer=l2(reg_param))(bn_pre_relu)
	l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
	l = Activation('relu')(l)
	
	# Dropout
	l = Dropout(0.3)(l)
	
	# Second layer
	conv2 = Convolution2D(nb_filter=nb_filters, nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
									subsample=(1, 1), W_regularizer=l2(reg_param))(l)
	
	
	if increase_dim:
		proj = Convolution2D(nb_filter=nb_filters, nb_row=1, nb_col=1, init='he_normal', border_mode='same', \
									subsample=(1, 1), W_regularizer=l2(reg_param), bias=False)(conv2)
		l = conv2 + proj
	elif first:
		proj = Convolution2D(nb_filter=nb_filters, nb_row=1, nb_col=1, init='he_normal', border_mode='same', \
									subsample=(1, 1), W_regularizer=l2(reg_param), bias=False)(conv2)
		l = conv2 + proj
	else:
		l = conv2 + x
	
	return l



inputs1 = tf.placeholder(dtype=tf.float32, shape=[None, image_size, image_size, 3])
inputs2 = tf.placeholder(dtype=tf.float32, shape=[None, image_size, image_size, 3])
labels = tf.placeholder(dtype=tf.float32, shape=[None, 1])

################################################ Run WideResNet ##################################################
# Merge inputs1 and inputs2
inputs = tf.concat(0, [inputs1, inputs2])

n_filters = [16, 16*wrn_k, 32*wrn_k]

# First layer
l = Convolution2D(nb_filter=n_filters[0], nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
									subsample=(1, 1), W_regularizer=l2(reg_param))(inputs)
l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
l = Activation('relu')(l)

# Second layer, a stack of residual blocks
l = residual_block(l, first=True, nb_filters=n_filters[1])
for _ in range(1, wrn_n):
	l = residual_block(l, nb_filters=n_filters[1])

# Third layer, a stack of residual blocks
l = residual_block(l, increase_dim=True, nb_filters=n_filters[2])
for _ in range(1, wrn_n+2):
	l = residual_block(l, nb_filters=n_filters[2])

l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
lstm_inputs = Activation('relu')(l)

# Divide lstm_inputs
lstm_inputs1 = lstm_inputs[:batch_size, :, :, :]
lstm_inputs2 = lstm_inputs[batch_size:, :, :, :]



############################################# Run the LSTM ########################################################
lstm_cell = tf.nn.rnn_cell.LSTMCell(hidden_layer_size, forget_bias = forget_bias)

with tf.variable_scope("attention"):
	W_g = tf.Variable(tf.random_normal([3, hidden_layer_size], stddev=0.01))

with tf.variable_scope("classifier"):
	W_c = tf.Variable(tf.random_normal([hidden_layer_size, 1], stddev=0.1))

#"""
h = lstm_cell.zero_state(batch_size, dtype=tf.float32)
c = tf.zeros((batch_size, hidden_layer_size))
im_patches = []
for j in range(2*num_glimpses):
	if j % 2 == 0:
		glimpse, patch = get_glimpse_batch(lstm_inputs1, tf.transpose(c), W_g, idx_to_show)
		glimpse = tf.reshape(glimpse, [batch_size, -1])
	else:
		glimpse, patch = get_glimpse_batch(lstm_inputs2, tf.transpose(c), W_g, idx_to_show)
		glimpse = tf.reshape(glimpse, [batch_size, -1])
	im_patches += [patch]
	with tf.variable_scope('mylstm') as scope:
		if j > 0:
			scope.reuse_variables()
		c, h = lstm_cell(glimpse, h)
im_patches = tf.stack(im_patches)

output = tf.nn.sigmoid(tf.matmul(c, W_c))
network_cost = -tf.reduce_mean(labels*tf.log(output) + (1.0-labels)*tf.log(1.0-output))
#"""

train_step = tf.train.AdamOptimizer(learning_rate).minimize(network_cost)
accuracy = tf.reduce_mean(tf.cast(tf.equal(tf.round(output), labels), tf.float32))



def train(images1, images2, is_similar):
	"""
		Train the network for verification task
	"""
	_, batch_cost = sess.run((train_step, network_cost), feed_dict={inputs1: images1, inputs2: images2, labels: is_similar, K.learning_phase(): 1})
	return batch_cost
	
	
def calculate_accuracy(images1, images2, is_similar, is_learning=0):
	"""
		Calculate accuracy for verification task
	"""
	#print sess.run(output, feed_dict={inputs1: images1, inputs2: images2, labels: is_similar})
	return sess.run(accuracy, feed_dict={inputs1: images1, inputs2: images2, labels: is_similar, K.learning_phase(): is_learning})
	
	

sess = tf.InteractiveSession()
saver = tf.train.Saver()
if os.path.isfile('./Model_Best/arc-model.meta'):
	saver.restore(sess, './Model_Best/arc-model')
else:
	sess.run(tf.global_variables_initializer())

K.set_session(sess)

# Useful controls
display_step = 100
save_step = 1000


# Read the Data
casia = CASIA('../Datasets/CASIA/CASIA-WebFace', (image_size, image_size))


####################################################### Training ##################################################

"""
for i in range(num_batches):
	images1, images2, is_similar = casia.get_data(batch_size)
	
	# Train on the batch
	train_cost = train(images1, images2, is_similar)
	
	if i % display_step == 0:		
		# Calculate accuracy on Validation Images
		images1, images2, is_similar = casia.get_data(batch_size, is_validation=True)
		val_accuracy = calculate_accuracy(images1, images2, is_similar)
		
		# Fetch some training images and calculate accuracy
		images1, images2, is_similar = casia.get_data(batch_size)
		train_accuracy = calculate_accuracy(images1, images2, is_similar)	
		print "Batch:", (i+1), "Validation Accuracy:", val_accuracy, "Train Cost:", train_cost, "Train Accuracy:", train_accuracy
	
	if i % save_step == 0:
		# Approximate validation cost over 
		val_cost = 0.0
		val_accuracy = 0.0
		for j in range(100):	
			images1, images2, is_similar = casia.get_data(batch_size, is_validation=True)
			v_cost, v_acc = sess.run((network_cost, accuracy), feed_dict={inputs1: images1, inputs2:images2, labels: is_similar, K.learning_phase(): 0})
			val_cost += v_cost
			val_accuracy += v_acc
		val_cost /= 100
		val_accuracy /= 100
		print 'Validation Set - Cost:', val_cost, 'Accuracy:', val_accuracy
		
		if val_cost < prev_best_val_cost:
			prev_best_val_cost = val_cost
			saver.save(sess, './Model_Best/best-model')	
		
		if val_cost - prev_val_cost >= eps:
			inc_cost_count += 1
			same_cost_count = 0
		elif val_cost - prev_val_cost <= -eps:
			inc_cost_count = 0
			same_cost_count = 0
		else:
			inc_cost_count = 0
			same_cost_count += 1		
		
		prev_val_cost = val_cost
		
		saver.save(sess, './Model/arc-model')
		
		if inc_cost_count >= inc_cost_threshold:
			print 'Ending Training: Validation Cost Increasing since', inc_cost_threshold*1000, 'iterations'
			break
		if same_cost_count >= same_cost_threshold:			
			print 'Ending Training: Validation Cost Same since', same_cost_threshold*1000, 'iterations'
			break
		

"""

####################################################### Testing ###################################################
# Calculate Test Accuracy
"""
num_trials = 10000
test_acc = 0.0
for i in range(num_trials):
	images1, images2, is_similar = casia.get_data_test(batch_size)
	test_accuracy = calculate_accuracy(images1, images2, is_similar)
	test_acc += test_accuracy
	print "Batch:", (i+1), "Test Accuracy:", test_accuracy

print "Accuracy:", (test_acc/num_trials)	
"""


############################################ Showing Attention Patches ############################################
# Get patches
#"""
for k in range(500):
	images1, images2, is_similar = casia.get_data_test(batch_size)
	image_patches = sess.run(im_patches, feed_dict={inputs1: images1, inputs2: images2, labels: is_similar, K.learning_phase(): 0})

	I1 = images1[idx_to_show, :, :, :]
	I2 = images2[idx_to_show, :, :, :]
	fig_axs = plt.subplots(2, num_glimpses)
	fig = fig_axs[0]
	axs = fig_axs[1:]
	axs = axs[0]
	for i in range(num_glimpses):
		ax = axs[0, i]
		ax.imshow(I1, cmap="Greys_r")
		ax.set_xticklabels([])
		ax.set_yticklabels([])
		ax.xaxis.grid(False) 
		ax.yaxis.grid(False) 
		ax.set_title(str(i + 1))
		x, y, w = 2*image_patches[2*i, 0], 2*image_patches[2*i, 1], 2*image_patches[2*i, 2]
		w *= attention_window_size
		print (x, y, w)
		x = x - w / 2.0
		y = 64 - y - w / 2.0
		rect = patches.Rectangle((x, y), w, w, linewidth=(2*w-1)/16, edgecolor='b', facecolor='none')
		ax.add_patch(rect)

		ax = axs[1, i]
		ax.imshow(I2, cmap="Greys_r")
		ax.set_xticklabels([])
		ax.set_yticklabels([])
		ax.xaxis.grid(False) 
		ax.yaxis.grid(False)
		x, y, w = 2*image_patches[2*i+1, 0], 2*image_patches[2*i+1, 1], 2*image_patches[2*i+1, 2]
		w *= attention_window_size
		print (x, y, w)
		x = x - w / 2.0
		y = 64 - y - w / 2.0
		rect = patches.Rectangle((x, y), w, w, linewidth=(2*w-1)/16, edgecolor='b', facecolor='none')
		ax.add_patch(rect)
	fig.savefig('./images_best/attention' + str(k) + '.png')
#"""
