import os
import numpy as np
import scipy.misc
import random

class MiniImageNet(object):
	"""
		Reads the MiniImageNet dataset
	"""
	
	def __init__(self, data_dir, image_size=(84, 84), test_data_fraction=0.2, seed=0.1):
		"""
			__init__(MiniImageNet, str, image_size) -> None
		"""
		self.image_size = image_size
		
		# Initialize the data
		self.num_training_examples = 0
		self.num_test_examples = 0
		self.test_data_fraction = test_data_fraction
		self.training_data = []
		self.test_data = []
		self.num_training_classes = 0
		self.num_test_classes = 0
		
		# Set random seed
		random.seed(seed)
		
		# Read the data
		self.read_data(data_dir)
	
	
	
	def read_data(self, data_dir):
		"""
			read_data(MiniImageNet, str) -> None
			Reads the training and test data
		"""
		
		# Read the directory names
		dirnames = []
		for _, dirnames, _ in os.walk(data_dir):
			break
			
		# Split into training and test data
		num_classes = len(dirnames)
		random.shuffle(dirnames)
		test_dirs = dirnames[:int(self.test_data_fraction*num_classes)]
		training_dirs = dirnames[int(self.test_data_fraction*num_classes):]
		print test_dirs
		print training_dirs
		
		# Read the training data
		i = 0
		num_classes_train = len(training_dirs)
		for class_dir in training_dirs:
			i += 1
			filenames = []
			for _, _, filenames in os.walk(os.path.join(data_dir, class_dir)):
				break
			
			num_files = len(filenames)
			self.num_training_examples += num_files
			data = np.zeros((num_files, self.image_size[0], self.image_size[1], 3))			
			j = 0		
			for filename in filenames:
				data[j, :, :, :] = self.load_image(os.path.join(data_dir, class_dir, filename))
				j += 1
				print '\rReading Training Class', i, 'of', num_classes_train, 'Progress:', 100*float(j)/num_files,
			
			self.training_data.append(data)
			print ''
		
		# Read the test data
		i = 0
		num_classes_test = len(test_dirs)
		for class_dir in test_dirs:
			i += 1
			filenames = []
			for _, _, filenames in os.walk(os.path.join(data_dir, class_dir)):
				break
			
			num_files = len(filenames)
			self.num_test_examples += num_files
			data = np.zeros((num_files, self.image_size[0], self.image_size[1], 3))			
			j = 0		
			for filename in filenames:
				data[j, :, :, :] = self.load_image(os.path.join(data_dir, class_dir, filename))
				j += 1
				print '\rReading Test Class', i, 'of', num_classes_test, 'Progress:', 100*float(j)/num_files,
			
			self.test_data.append(data)
			print ''
			
		# Record statistics
		self.num_training_classes = len(self.training_data)
		self.num_test_classes = len(self.test_data)
			
			
		
	def load_image(self, filename):
		"""
			load_image(str) -> ndarray
			Reads the image from the specified file and returns an image of size = self.image_size
		"""
		# Read the image
		im = scipy.misc.imread(filename)
				
		# Resize the image
		im = scipy.misc.imresize(im, self.image_size)
		
		# Bring the image in range [0, 1]
		im = np.ndarray.astype(im, np.float32) / 255
		
		return im
			
	
	
	def get_data(self, batch_size):
		"""
			get_data(MiniImageNet, int) -> images1, images2, is_similar
			Returns the data to the client
		"""
		
		# Initialize the data
		images1 = np.zeros((batch_size, self.image_size[0], self.image_size[1], 3))
		images2 = np.zeros((batch_size, self.image_size[0], self.image_size[1], 3))
		is_similar = np.zeros((batch_size, 1))
		
		for i in range(batch_size):
			im1 = None
			im2 = None
			if i % 2 == 0:
				im1, im2 = self.read_similar_images()
				is_similar[i] = 1.0
			else:
				im1, im2 = self.read_different_images()
				is_similar[i] = 0.0
			
			images1[i, :, :, :] = im1
			images2[i, :, :, :] = im2
		
		return (images1, images2, is_similar)
		
		
	
	def get_data_test(self, batch_size):
		"""
			get_data(MiniImageNet, int) -> images1, images2, is_similar
			Returns the data to the client
		"""
		
		# Initialize the data
		images1 = np.zeros((batch_size, self.image_size[0], self.image_size[1], 3))
		images2 = np.zeros((batch_size, self.image_size[0], self.image_size[1], 3))
		is_similar = np.zeros((batch_size, 1))
		
		for i in range(batch_size):
			im1 = None
			im2 = None
			if i % 2 == 0:
				im1, im2 = self.read_similar_images_test()
				is_similar[i] = 1.0
			else:
				im1, im2 = self.read_different_images_test()
				is_similar[i] = 0.0
			
			images1[i, :, :, :] = im1
			images2[i, :, :, :] = im2
		
		return (images1, images2, is_similar)
	
	
	
	def read_similar_images_test(self):
		"""
			read_similar_images_test(MiniImageNet) -> image1, image2
			Reads two distinct images from a single test class
		"""
		# Randomly select a class to read
		idx = random.randint(0, self.num_test_classes-1)
			
		# Randomly select two images from this class
		num_images = np.shape(self.test_data[idx])[0]
		image_idx = random.sample(range(num_images), 2)
		
		return (self.test_data[idx][image_idx[0], :, :, :], self.test_data[idx][image_idx[1], :, :, :])	
	
	
	
	def read_similar_images(self):
		"""
			read_similar_images(MiniImageNet) -> image1, image2
			Reads two distinct images from a single training class
		"""
		# Randomly select a class to read
		idx = random.randint(0, self.num_training_classes-1)
			
		# Randomly select two images from this class
		num_images = np.shape(self.training_data[idx])[0]
		image_idx = random.sample(range(num_images), 2)
		
		return (self.training_data[idx][image_idx[0], :, :, :], self.training_data[idx][image_idx[1], :, :, :])
		
	
	def read_different_images(self):
		"""
			read_different_images(MiniImageNet) -> image1, image2
			Reads two images from different training classes
		"""
		# Randomly select a class to read
		idx = random.sample(range(self.num_training_classes), 2)
		
		# Randomly select an image from each of the class
		num_images = np.shape(self.training_data[idx[0]])[0]
		image_idx0 = random.randint(0, num_images-1)
		num_images = np.shape(self.training_data[idx[1]])[0]
		image_idx1 = random.randint(0, num_images-1)
		
		return (self.training_data[idx[0]][image_idx0, :, :, :], self.training_data[idx[1]][image_idx1, :, :, :])
		
	
	def read_different_images_test(self):
		"""
			read_different_images_test(MiniImageNet) -> image1, image2
			Reads two images from different test classes
		"""
		# Randomly select a class to read
		idx = random.sample(range(self.num_test_classes), 2)
		
		# Randomly select an image from each of the class
		num_images = np.shape(self.test_data[idx[0]])[0]
		image_idx0 = random.randint(0, num_images-1)
		num_images = np.shape(self.test_data[idx[1]])[0]
		image_idx1 = random.randint(0, num_images-1)
		
		return (self.test_data[idx[0]][image_idx0, :, :, :], self.test_data[idx[1]][image_idx1, :, :, :])
	
	
	
	def get_one_shot_data(self, n):
		"""
			get_one_shot_data(MiniImageNet, int) -> training_images, test_image, correct_idx
			Returns the data for n-way one shot classification task
		"""
		# Initialize the data
		training_images = np.zeros((n, self.image_size[0], self.image_size[1], 3))
		test_image = np.zeros((self.image_size[0], self.image_size[1], 3))
		
		# Put one random example from each class into training_images
		chosen_idx = []
		# Select random classes
		classes = random.sample(range(self.num_test_classes), n)
		k = 0
		for i in classes:
			# Select a random example
			num_images = np.shape(self.test_data[i])[0]
			idx = random.randint(0, num_images-1)
			chosen_idx.append(idx)	# Record which index from this class was chosen for training
			training_images[k, :, :, :] = self.test_data[i][idx, :, :, :]
			k += 1
		
		# Choose a random test_image
		class_idx = random.choice(classes)
		idx = classes.index(class_idx)
		num_images = np.shape(self.test_data[class_idx])[0]
		image_idx = random.randint(0, num_images-1)
		while image_idx == chosen_idx[idx]:
			image_idx = random.randint(0, num_images-1)
		test_image = self.test_data[class_idx][image_idx, :, :, :]
		
		return (training_images, test_image, idx)	
		
			
			
	def get_num_training_examples(self):
		"""
			get_num_training_examples(MiniImageNet) -> int
			Returns the number of training examples
		"""
		return self.num_training_examples
		
		
	def get_num_test_examples(self):
		"""
			get_num_test_examples(MiniImageNet) -> int
			Returns the number of test examples
		"""
		return self.num_test_examples
	
	
		
if __name__ == '__main__':
	provider = MiniImageNet('./Mini-ImageNet')
	images1, images2, is_similar = provider.get_data(128)
	training_images, test_image, correct_idx = provider.get_one_shot_data()
